
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>


enum {
//	_xinfreq = 20_000_000,
	_xtlfreq = 20_000_000,
	_clkfreq = 40_000_000,

	DOWNLOAD_BAUD = 230_400,
	DEBUG_BAUD = DOWNLOAD_BAUD,

	HEAPSIZE = 8800,

//
// Targetted at controlling Roger Loh's 4-bit SD add-on board
//
	SD_BASE_PIN = 16,

	PIN_DAT    = SD_BASE_PIN+0,    // MISO // DAT0
	PIN_MISO   = SD_BASE_PIN+0,    // MISO // DAT0
	PIN_CS     = SD_BASE_PIN+3,    // CS   // DAT3
	PIN_CMD    = SD_BASE_PIN+4,    // MOSI // CMD
	PIN_MOSI   = SD_BASE_PIN+4,    // MOSI // CMD
	PIN_CLK    = SD_BASE_PIN+5,    // CLK

	MAX_BUFFER_SIZE = (100 * 1024)
};

static int  mounted = 0;
static int32_t  xoro32state;


static void  randfill( uint32_t *addr, size_t size )
{
    int32_t  state = xoro32state;
    size >>= 2;

    __asm volatile {    // "volatile" enforces unoptimised use of FCACHE
		wrfast	#0, addr
		rep	@.rend, size
		xoro32	state
		mov	pb, 0-0
		wflong	pb
.rend
    }
    xoro32state = state;
}


static size_t  randcmp( const uint32_t *addr, size_t size )
{
    int32_t  state = xoro32state;
    size_t  pass = size >> 2;

    __asm volatile {    // "volatile" enforces unoptimised use of FCACHE
		rdfast	#0, addr
.loop
		rflong	pb
		xoro32	state
		cmp	pb, 0-0  wz
	if_z	djnz	pass, #.loop
    }
    xoro32state = state;

    return  size - (pass << 2);
}


static size_t  fgetsize( const char *filename )
{
    struct stat  st;

    errno = 0;
    if(stat(filename, &st))
        exit(1);
    return  st.st_size;  // filesize in bytes
}


static void  mountall( void )
{
    DIR  *fh;

    umount("/sd1");
    umount("/sd2");

    mount("/sd1", _vfs_open_sdcard());
    errno = 0;
    fh = opendir("/sd1/.");  // doesn't set errno on success
    perror("mount sd1");
    if(!fh)  exit(1);    // requires #include <errno.h>
    closedir(fh);
/*
    mount("/sd2", _vfs_open_sdcardx(PIN_CLK, PIN_CS, PIN_MOSI, PIN_MISO));
    errno = 0;
    fh = opendir("/sd2/.");  // doesn't set errno on success
    perror("mount sd2");
    if(!fh)  exit(1);    // requires #include <errno.h>
    closedir(fh);
*/
    mount("/sd2", _vfs_open_sdcard4b(PIN_CLK, PIN_CMD, PIN_DAT));
    errno = 0;
    fh = opendir("/sd2/.");  // doesn't set errno on success
    perror("mount sd2");
    if(!fh)  exit(1);    // requires #include <errno.h>
    closedir(fh);

    mounted = 1;
}


static void  rmfile( const char *filename )
{
    if(!mounted)
        mountall();

    printf("  deleting file %s ...", filename);
    errno = 0;
    remove(filename);  // ignore any errors
    perror("  deleted");
}


static void  writefile( const char *filename, size_t filesize )
{
    FILE  *fh;
    size_t  bufsize, count = 0;
    uint32_t  ticks, ticke, msecs = 0, onesecond, *data;

    if(!mounted)
        mountall();

    if(filesize / MAX_BUFFER_SIZE)  bufsize = MAX_BUFFER_SIZE;
    else  bufsize = filesize;

    if(data = __builtin_alloca(bufsize))    // auto-frees upon return
        printf("write buffer addr = 0x%x, size = %d\n", data, bufsize);
    else {
        printf("malloc() failed!\n");
        exit(1);
    }

    onesecond = _clockfreq();
    ticks = _cnt();

    errno = 0;
    fh = fopen(filename, "w");
    if(!fh) {
        perror("fopen() for writing failed\n");
        exit(1);
    }

    do {
        randfill(data, bufsize);
        if(fwrite(data, bufsize, 1, fh))  count += bufsize;
        else  exit(1);
        ticke = _cnt();
        while(ticke - ticks >= onesecond)  {
            ticks += onesecond;
            msecs += 1000;
        }
    } while(count < filesize);

    fclose(fh);
    ticke = _cnt();
    msecs += _muldiv64(ticke - ticks, 1000, _clockfreq());  // ticks to milliseconds
    printf("Written %d of %d bytes at %d kB/s\n", count, filesize,
        (_muldiv64(count, 1000, msecs) + 512) >> 10);

}


static void  copyfile( const char *filename1, const char *filename2 )
{
    FILE  *fh1, *fh2;
    size_t  filesize, bufsize, count = 0;
    uint32_t  ticks, ticke, msecs = 0, onesecond, *data;

    if(!mounted)
        mountall();

    filesize = fgetsize(filename1);
    printf("copy file length = %d\n", filesize);

    if(filesize / MAX_BUFFER_SIZE)  bufsize = MAX_BUFFER_SIZE;
    else  bufsize = filesize;

    if(data = __builtin_alloca(bufsize))    // auto-frees upon return
        printf("copy buffer addr = 0x%x, size = %d\n", data, bufsize);
    else {
        printf("malloc() failed!\n");
        exit(1);
    }
    memset(data, 0, bufsize);  // erase any echoes

    onesecond = _clockfreq();
    ticks = _cnt();

    errno = 0;
    fh1 = fopen(filename1, "r");
    if(!fh1) {
        perror("fopen() for reading failed\n");
        exit(1);
    }

    fh2 = fopen(filename2, "w");
    if(!fh2) {
        printf("fopen() for writing failed\n");
        exit(1);
    }

    do {
        if(!fread(data, bufsize, 1, fh1))  exit(1);
        ticke = _cnt();
        while(ticke - ticks >= onesecond)  {
            ticks += onesecond;
            msecs += 1000;
        }
        if(fwrite(data, bufsize, 1, fh2))  count += bufsize;
        else  exit(1);
        ticke = _cnt();
        while(ticke - ticks >= onesecond)  {
            ticks += onesecond;
            msecs += 1000;
        }
    } while(count < filesize);

    fclose(fh1);
    fclose(fh2);
    ticke = _cnt();
    msecs += _muldiv64(ticke - ticks, 1000, _clockfreq());  // ticks to milliseconds
    printf("Copied %d of %d bytes at %d kB/s\n", count, filesize,
        (_muldiv64(count, 1000, msecs) + 512) >> 10);

}


static void  checkfile( const char *filename )
{
    FILE  *fh;
    size_t  filesize, bufsize, count = 0;
    uint32_t  ticks, ticke, tickf, msecs = 0, onesecond, *data;
    int  passed = 1;

    if(!mounted)
        mountall();

    filesize = fgetsize(filename);
    printf("read back file length = %d\n", filesize);

    if(filesize / MAX_BUFFER_SIZE)  bufsize = MAX_BUFFER_SIZE;
    else  bufsize = filesize;

    if(data = __builtin_alloca(bufsize))    // auto-frees upon return
        printf("read back buffer addr = 0x%x, size = %d\n", data, bufsize);
    else {
        printf("malloc() failed!\n");
        exit(1);
    }
    memset(data, 0, bufsize);  // erase any echoes

    onesecond = _clockfreq();
    ticks = _cnt();
    tickf = ticks;

    errno = 0;
    fh = fopen(filename, "r");
    if(!fh) {
        perror("fopen() for read back failed\n");
        exit(1);
    }

    do {
        if(!fread(data, bufsize, 1, fh))  exit(1);
        ticke = _cnt();
        while(ticke - tickf >= onesecond)  {
            tickf += onesecond;
            msecs += 1000;
        }
        if(bufsize == randcmp(data, bufsize))  count += bufsize;
        else {
            passed = 0;
            break;
        }
    } while(count < filesize);

    fclose(fh);
    ticke = _cnt();
/*    ticks = ticke - ticks;
    printf("Read back %d of %d bytes at %d kB/s\n", count, filesize,
        (_muldiv64(count, _clockfreq(), ticks) + 512) >> 10);
*/
    msecs += _muldiv64(ticke - tickf, 1000, _clockfreq());  // ticks to milliseconds
//    printf("ticks = %d, msecs = %d\n", ticks, msecs);
    printf("Read back %d of %d bytes at %d kB/s\n", count, filesize,
        (_muldiv64(count, 1000, msecs) + 512) >> 10);

    if( passed )  puts("Compare match.  :)");
    else  puts("Compare mismatch!");
}


//===================================================================================

static void  shutdown( void )
{
    printf("\nClean-up ...\n errno = %d: %s\n", errno, strerror(errno));

    umount("/sd1");
    umount("/sd2");

    _waitms(500);
    _clkset(0, 24_000_000);  // cool running
}


const char filepath1[] = "/sd1/copy1.bin";
const char filepath2[] = "/sd2/copy1.bin";


void  main( void )
{
    uint32_t  seed;
    int  i;

    printf("\n   clkfreq = %d   clkmode = 0x%x\n\n", _clockfreq(), _clockmode());
    atexit(shutdown);

    for(i = 0; i < 3; i++)
    {
        rmfile(filepath1);
        rmfile(filepath2);
        seed = _rnd();
        printf("data seed = 0x%08x\n", seed);
        xoro32state = seed;
        writefile(filepath1, 4_400_000);
        xoro32state = seed;
        checkfile(filepath1);
        copyfile(filepath1, filepath2);
        xoro32state = seed;
        checkfile(filepath2);
    }

    exit(0);
}
